/*  eqskin.c
 *  simple skin support in EQ plugin
 *  (c) 2003 by Simone Karin Lehmann <simone@lisanet.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "eqskin.h"
#include "eq.h"
#include "bmp.h"

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <ctype.h>
#include <dirent.h>
#ifdef HAVE_FTS_H
#include <fts.h>
#endif

// include all xpm for the skins
#include "defskin/eqmain31.xpm"
#include "defskin/eqmain25.xpm"
#include "defskin/eqmain15.xpm"
#include "defskin/eqmain10.xpm"
#include "defskin/eqmain10_orig.xpm"
#include "defskin/eqcontrols.xpm"
#include "defskin/eqboost.xpm"
#include "defskin/eqoverlay.xpm"

static gint skin_width;
static gint skin_goffset;

static GdkPixmap *skin_controls, *skin_boost, *skin_overlay;

char *tmpdir;

GdkPixmap *eq_load_default_skin(GdkWindow  *window)
{
	char **skin_name = NULL;

	if (eqcfg.band_num == 31)
	{
		skin_name = eqmain31_xpm;
	}
	else if (eqcfg.band_num == 15)
	{
		skin_name = eqmain15_xpm;
	}
	else if (eqcfg.band_num == 25)
	{
		skin_name = eqmain25_xpm;
	}
	else if (eqcfg.band_num == 10)
	{
		if (eqcfg.use_xmms_original_freqs)
			skin_name = eqmain10_orig_xpm;
		else
			skin_name = eqmain10_xpm;
	}

	skin_controls = gdk_pixmap_create_from_xpm_d(window, NULL, NULL, eqcontrols_xpm);
	skin_boost = gdk_pixmap_create_from_xpm_d(window, NULL, NULL, eqboost_xpm);
	skin_overlay = gdk_pixmap_create_from_xpm_d(window, NULL, NULL, eqoverlay_xpm);

	return gdk_pixmap_create_from_xpm_d(window, NULL, NULL, skin_name);

}

GdkPixmap *eq_load_skin(GdkWindow  *window)
{
	char *filename = NULL;
	gchar *f;
	GdkPixmap *pix;

	if (eqcfg.band_num == 31)
	{
		filename = "eqmain31.bmp";
	}
	else if (eqcfg.band_num == 15)
	{
		filename = "eqmain15.bmp";
	}
	else if (eqcfg.band_num == 25)
	{
		filename = "eqmain25.bmp";
	}
	else if (eqcfg.band_num == 10)
	{
		if (eqcfg.use_xmms_original_freqs)
			filename = "eqmain10_orig.bmp";
		else
			filename = "eqmain10.bmp";
	}

	f = g_strconcat(tmpdir, "/", filename, NULL);
	pix = read_bitmap(window, f);
	g_free(f);

	f = g_strconcat(tmpdir, "/eqcontrols.bmp", NULL);
	skin_controls = read_bitmap(window,f);
	g_free(f);

	f = g_strconcat(tmpdir, "/eqboost.bmp", NULL);
	skin_boost = read_bitmap(window,f);
	g_free(f);

	f = g_strconcat(tmpdir, "/eqoverlay.bmp", NULL);
	skin_overlay = read_bitmap(window,f);
	g_free(f);

	return pix;
}


char *eq_decompress_skin(const char* path)
{
	char *tmp = NULL, *tempdir, *unzip, *tar, *ending;

	unzip = getenv("UNZIPCMD");
	if (!unzip)
		unzip = "unzip";
	tar = getenv("TARCMD");
	if (!tar)
		tar = "tar";

	if ((ending = strrchr(path, '.')) == NULL)
		return NULL;

	tempdir = g_strconcat(g_get_tmp_dir(), "/eqskin", NULL);
	if (mkdir(tempdir,0700))
    {
		g_free(tempdir);
		g_message("Failed to create temporary directory: %s.  Unable to load skin.", strerror(errno));
		return NULL;
	}

	if (!strcasecmp(ending, ".zip") || !strcasecmp(ending, ".wsz"))
		tmp = g_strdup_printf("%s >/dev/null -o -j \"%s\" -d %s", unzip, path, tempdir);
	if (!strcasecmp(ending, ".tgz") || !strcasecmp(ending, ".gz"))
		tmp = g_strdup_printf("%s >/dev/null xzf \"%s\" -C %s", tar, path, tempdir);
	if (!strcasecmp(ending, ".bz2"))
		tmp = g_strdup_printf("bzip2 -dc \"%s\" | %s >/dev/null xf - -C %s", path, tar, tempdir);
	if (!strcasecmp(ending, ".tar"))
		tmp = g_strdup_printf("%s >/dev/null xf \"%s\" -C %s", tar, path, tempdir);

	system(tmp);
	g_free(tmp);
	return tempdir;
}




GdkPixmap *eq_create_skin_bg_pixmap(GdkWindow  *window)
{
	GdkPixmap *bg, *pix;
	GdkGC *gc;
	gint h, over;

// load background skin

	if (!strcmp(eqcfg.skin, "default"))
    {
		pix = eq_load_default_skin(window);
	}
	else
	{
		tmpdir = eq_decompress_skin(eqcfg.skin);
		pix = eq_load_skin(window);
		del_directory(tmpdir);
		g_free(tmpdir);
	}

	gdk_window_get_size(pix, &skin_width, &h);

	skin_goffset = (skin_width/2-113)/2;
	if (eqcfg.band_num == 10)
		skin_goffset += 5;
	if (!eqcfg.use_independent_channels)
		skin_width /= 2;

	// create background
	bg = gdk_pixmap_new(NULL, skin_width, 116, gdk_rgb_get_visual()->depth);
	gc = gdk_gc_new(window);
	// paint skin on background and set it for the window
	gdk_draw_pixmap(bg, gc, pix, 0, 0, 0, 0, skin_width, 116);
	// paint the overlay if we use only hvae no indep. channels
	over = 0;
	if (!eqcfg.use_independent_channels)
	{
		if (eqcfg.band_num == 25)
			over = 37;
		gdk_draw_pixmap(bg, gc, skin_overlay, 0, over, skin_width - 60, 0, 60, 36);
	}

	gdk_window_clear(window);
	gdk_window_set_back_pixmap(window, bg, 0);

	return bg;

}

GdkPixmap *eq_skin_controls(void)
{
	return skin_controls;
}

GdkPixmap *eq_skin_boost(void)
{
	return skin_boost;
}


gint eq_skin_width(void)
{
	return skin_width;
}

gint eq_skin_goffset(void)
{
	return skin_goffset;
}


// ************ taken from xmms *******************

void del_directory(const char *dirname)
{
#ifdef HAVE_FTS_H
	char * const argv[2] = { (char *) dirname, NULL };
	FTS *fts;
	FTSENT *p;

	fts = fts_open(argv, FTS_PHYSICAL, (int (*)())NULL);
	while ((p = fts_read(fts)) != NULL) {
		switch (p->fts_info) {
		case FTS_D:
			break;
		case FTS_DNR:
		case FTS_ERR:
			break;
		case FTS_DP:
			rmdir(p->fts_accpath);
			break;
		default:
			unlink(p->fts_accpath);
			break;
		}
	}
#else /* !HAVE_FTS_H */
	DIR *dir;
	struct dirent *dirent;
	char *file;

	if ((dir = opendir(dirname)) != NULL)
	{
		while ((dirent = readdir(dir)) != NULL)
		{
			if (strcmp(dirent->d_name, ".") && strcmp(dirent->d_name, ".."))
			{
				file = g_strdup_printf("%s/%s", dirname, dirent->d_name);
				if (unlink(file) == -1)
					if (errno == EISDIR)
						del_directory(file);
				g_free(file);
			}
		}
		closedir(dir);
	}
	rmdir(dirname);
#endif /* !HAVE_FTS_H */
}




